from __future__ import annotations
import json
from typing import Optional
import numpy as np
import zmq
from qtpy.QtCore import QObject, Signal
from ._utils import default_endpoint


class ZMQImageListener(QObject):
    """Background worker that pulls image frames over ZeroMQ.

    Emits `received(array, meta)` whenever an image arrives.
    """

    received = Signal(object, dict)
    status = Signal(str)
    error = Signal(str)

    def __init__(self, endpoint: Optional[str] = None, high_water_mark: int = 10):
        super().__init__()
        self._endpoint = endpoint or default_endpoint()
        self._hwm = high_water_mark
        self._ctx: Optional[zmq.Context] = None
        self._sock: Optional[zmq.Socket] = None
        self._running = False

    def start(self):
        try:
            self._ctx = zmq.Context.instance()
            self._sock = self._ctx.socket(zmq.PULL)
            self._sock.setsockopt(zmq.RCVHWM, self._hwm)
            self._sock.setsockopt(zmq.LINGER, 0)

            # Bind as a local server that senders connect to
            self._sock.bind(self._endpoint)
            self._running = True
            self.status.emit(f"Listening on {self._endpoint}")

            poller = zmq.Poller()
            poller.register(self._sock, zmq.POLLIN)

            while self._running:
                events = dict(poller.poll(timeout=100))  # 100 ms tick for responsiveness
                if self._sock in events:
                    try:
                        header_b, payload = self._sock.recv_multipart(flags=zmq.NOBLOCK)
                        meta = json.loads(header_b.decode("utf-8"))
                        arr = _from_bytes(payload, meta)
                        self.received.emit(arr, meta)
                    except zmq.Again:
                        pass
                    except Exception as e:  # noqa: BLE001
                        self.error.emit(repr(e))
        except Exception as e:  # noqa: BLE001
            self.error.emit(repr(e))
        finally:
            self._teardown()

    def stop(self):
        self._running = False
        self.status.emit("Stopping listener…")

    def _teardown(self):
        try:
            if self._sock is not None:
                self._sock.close(0)
        finally:
            self._sock = None
            self._ctx = None
            self.status.emit("Listener stopped.")


def bind_endpoint_for_public(endpoint: str) -> str:
    """Convert a TCP endpoint to bind on all interfaces."""
    if not endpoint.startswith("tcp://"):
        return endpoint
    try:
        host_port = endpoint[len("tcp://"):]
        host, port = host_port.rsplit(":", 1)
        port_int = int(port)
    except Exception:
        return endpoint
    if host in ("*", "0.0.0.0"):
        return endpoint
    return f"tcp://*:{port_int}"


def _from_bytes(buf: bytes, meta: dict) -> np.ndarray:
    """Reconstruct ndarray from raw bytes and metadata.

    Expected `meta` keys: shape, dtype, order ("C"/"F").
    Optional: For clarity, we also pass through viewer kwargs like name/colormap/etc.
    """
    shape = tuple(meta["shape"])  # e.g. [Z, Y, X] or [Y, X]
    dtype = np.dtype(meta["dtype"])  # e.g. "float32"
    order = meta.get("order", "C")

    arr = np.frombuffer(buf, dtype=dtype)
    arr = arr.reshape(shape, order=order)
    return arr



from __future__ import annotations
import socket

DEFAULT_TCP_PORT = 5556


def default_endpoint(public: bool = False) -> str:
    if public:
        return f"tcp://{_preferred_ip()}:{DEFAULT_TCP_PORT}"
    # TCP everywhere for simplicity
    return f"tcp://127.0.0.1:{DEFAULT_TCP_PORT}"


def _preferred_ip() -> str:
    """Best-effort guess of a non-loopback IPv4 address for sharing endpoints."""
    try:
        infos = socket.getaddrinfo(socket.gethostname(), None, family=socket.AF_INET)
        for _, _, _, _, (addr, *_rest) in infos:
            if not addr.startswith("127."):
                return addr
    except Exception:
        pass
    return "127.0.0.1"



import numpy as np

from napari_stream._receiver_widget import (
    ReceiverWidget,
    bind_endpoint_for_public,
    receiver_widget,
)


def test_bind_endpoint_for_public():
    assert bind_endpoint_for_public("tcp://127.0.0.1:5556") == "tcp://*:5556"
    assert bind_endpoint_for_public("tcp://*:5556") == "tcp://*:5556"
    assert bind_endpoint_for_public("ipc:///tmp/napari_stream.sock") == "ipc:///tmp/napari_stream.sock"


def test_receiver_widget_uses_existing_viewer(make_napari_viewer):
    viewer = make_napari_viewer()
    widget = receiver_widget(viewer)
    assert isinstance(widget, ReceiverWidget)
    assert widget.viewer is viewer


def test_autocontrast_updates_layer_limits(make_napari_viewer):
    viewer = make_napari_viewer()
    widget = receiver_widget(viewer)

    # simulate a received grayscale image without contrast_limits
    arr = np.random.rand(32, 32).astype(np.float32)
    meta = {"name": "foo", "rgb": False}
    widget._on_received(arr, meta)

    layer = viewer.layers["foo"]
    assert layer.contrast_limits is not None



from __future__ import annotations
import traceback
import shutil
import subprocess
from typing import Optional, Tuple
import numpy as np
from qtpy.QtCore import QThread
from qtpy.QtGui import QGuiApplication
from qtpy.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QLineEdit,
    QPushButton,
    QLabel,
    QHBoxLayout,
    QCheckBox,
    QRadioButton,
    QButtonGroup,
    QGroupBox,
)

from ._listener import ZMQImageListener, bind_endpoint_for_public
from ._utils import DEFAULT_TCP_PORT, default_endpoint

try:
    from napari.types import ImageData
    from napari import Viewer
except Exception:  # pragma: no cover
    Viewer = object  # type: ignore[misc,assignment]
    ImageData = np.ndarray  # type: ignore[assignment]


class ReceiverWidget(QWidget):
    def __init__(self, napari_viewer: Viewer):
        super().__init__()
        self.viewer = napari_viewer

        self._thread: Optional[QThread] = None
        self._worker: Optional[ZMQImageListener] = None

        # Mode-specific endpoints
        self._endpoint_local = f"tcp://127.0.0.1:{DEFAULT_TCP_PORT}"
        self._endpoint_private = default_endpoint(public=True)
        self._endpoint_tunnel = ""

        self._tunnel_proc: Optional[subprocess.Popen] = None
        self._last_mode = "local"

        self.endpoint_edit = QLineEdit(self._endpoint_local)
        self.status_label = QLabel("Idle")
        self.mode_local = QRadioButton("This computer only")
        self.mode_tunnel = QRadioButton("Remote host via SSH")
        self.mode_private = QRadioButton("Local network")
        self.mode_local.setChecked(True)
        self.mode_group = QButtonGroup(self)
        for btn in (self.mode_local, self.mode_tunnel, self.mode_private):
            self.mode_group.addButton(btn)
        self.mode_local.setToolTip("Listen on 127.0.0.1 only. Use when sender runs on this machine.")
        self.mode_private.setToolTip("Bind to your LAN/VPN IP so other machines can connect directly.\nShare the shown tcp://<ip>:port endpoint with the sender (e.g., set NAPARI_STREAM_ENDPOINT).\nNote: some corporate/VPN setups block inbound connections; if this fails, use the SSH tunnel mode.")
        self.mode_tunnel.setToolTip("Create an SSH reverse tunnel to a remote host. Enter SSH target as user@host[#port]; sender connects to its local tcp://127.0.0.1:<port>.")

        self.autocontrast = QCheckBox("Auto-contrast on new images")
        self.autocontrast.setChecked(True)
        self.ignore_affine = QCheckBox("Ignore image affine")
        self.ignore_affine.setToolTip("Ignore affine metadata when adding images. Use if napari errors on certain affines.")
        self.verbose = QCheckBox("Verbose")

        self.btn_run = QPushButton("Start")
        self.btn_copy = QPushButton("Copy Endpoint")

        top = QVBoxLayout(self)
        top.addWidget(QLabel("Endpoint:"))
        top.addWidget(self.endpoint_edit)
        top.addWidget(self.btn_copy)
        mode_box = QGroupBox("Connection mode")
        mode_layout = QVBoxLayout(mode_box)
        mode_layout.addWidget(self.mode_local)
        mode_layout.addWidget(self.mode_tunnel)
        mode_layout.addWidget(self.mode_private)
        top.addWidget(mode_box)
        top.addWidget(self.autocontrast)
        top.addWidget(self.ignore_affine)
        top.addWidget(self.verbose)
        top.addWidget(self.status_label)
        row2 = QHBoxLayout()
        row2.addWidget(self.btn_run)
        top.addLayout(row2)

        self.btn_run.clicked.connect(self._on_toggle_clicked)
        self.btn_copy.clicked.connect(self._copy_endpoint)
        self.mode_group.buttonClicked.connect(self._on_mode_changed)
        self.destroyed.connect(lambda *_: self._stop_tunnel())
        app = QGuiApplication.instance()
        if app is not None:
            app.aboutToQuit.connect(lambda *_: self._stop_tunnel())

    def _on_toggle_clicked(self):
        if self._is_running():
            self._on_stop()
        else:
            self._on_start()

    def _on_start(self):
        if self._is_running():
            return
        mode = self._current_mode()
        user_entry = self.endpoint_edit.text().strip()
        endpoint = user_entry or self._endpoint_for_mode(mode)
        if mode == "local":
            endpoint = self._ensure_local_endpoint(endpoint, mode="local")
        elif mode == "private":
            endpoint = self._ensure_private_endpoint(endpoint)
        elif mode == "tunnel":
            ssh_target, ssh_port = self._parse_ssh_target(user_entry)
            if not ssh_target:
                self.status_label.setText("Enter SSH target for reverse tunnel.")
                return
            endpoint = self._ensure_local_endpoint(
                self._endpoint_tunnel or f"tcp://127.0.0.1:{DEFAULT_TCP_PORT}",
                mode="tunnel",
                update_field=False,
            )
            self._store_endpoint_for_mode("tunnel", user_entry)
        else:
            self.status_label.setText("Unknown mode.")
            return

        self._thread = QThread()
        self._worker = ZMQImageListener(endpoint)
        self._worker.moveToThread(self._thread)

        self._thread.started.connect(self._worker.start)
        self._worker.received.connect(self._on_received)
        self._worker.status.connect(self.status_label.setText)
        self._worker.error.connect(self._on_error)

        self.btn_run.setText("Stop")
        self.btn_run.setEnabled(True)
        self._thread.start()

        if mode == "tunnel":
            self._log_verbose(f"Starting reverse tunnel to {ssh_target}:{ssh_port} (local {endpoint})")
            if not self._start_reverse_tunnel(ssh_target, ssh_port, endpoint):
                self._on_stop()

    def _on_stop(self):
        self._stop_tunnel()
        if self._worker is not None:
            self._worker.stop()
        if self._thread is not None:
            self._thread.quit()
            self._thread.wait()
        self._thread = None
        self._worker = None
        self.btn_run.setText("Start")
        self.btn_run.setEnabled(True)

    def _copy_endpoint(self):
        endpoint = self.endpoint_edit.text().strip()
        QGuiApplication.clipboard().setText(endpoint)

    def _on_mode_changed(self, *_):
        # Stop if running to allow reconfiguration
        if self._is_running():
            self._on_stop()
        # Persist current text to the previous mode slot
        prev_mode = getattr(self, "_last_mode", None)
        if prev_mode:
            self._store_endpoint_for_mode(prev_mode, self.endpoint_edit.text().strip())
        mode = self._current_mode()
        self.endpoint_edit.setPlaceholderText("user@remote or ssh-alias[#port]" if mode == "tunnel" else "")
        self.endpoint_edit.setText(self._endpoint_for_mode(mode))
        self.btn_copy.setEnabled(mode != "tunnel")
        self._last_mode = mode
        self.status_label.setText("Idle")
        self._log_verbose(f"Switched mode to {mode}")

    def _current_mode(self) -> str:
        if self.mode_private.isChecked():
            return "private"
        if self.mode_tunnel.isChecked():
            return "tunnel"
        return "local"

    def _endpoint_for_mode(self, mode: str) -> str:
        if mode == "private":
            return self._endpoint_private
        if mode == "tunnel":
            # Show blank to prompt SSH target entry; fall back to stored value if present
            return "" if not self._endpoint_tunnel else self._endpoint_tunnel
        return self._endpoint_local

    def _store_endpoint_for_mode(self, mode: str, endpoint: str) -> None:
        if mode == "private":
            self._endpoint_private = endpoint or self._endpoint_private
        elif mode == "tunnel":
            self._endpoint_tunnel = endpoint or self._endpoint_tunnel
        else:
            self._endpoint_local = endpoint or self._endpoint_local

    def _ensure_local_endpoint(self, endpoint: str, *, mode: str = "local", update_field: bool = True) -> str:
        port = self._extract_port(endpoint) or DEFAULT_TCP_PORT
        value = f"tcp://127.0.0.1:{port}"
        if update_field:
            self.endpoint_edit.setText(value)
            self._store_endpoint_for_mode(mode, value)
        return value

    def _ensure_private_endpoint(self, endpoint: str) -> str:
        # Prefer a shareable IP, then bind on all interfaces.
        if not endpoint.startswith("tcp://"):
            endpoint = self._endpoint_private or default_endpoint(public=True)
        if endpoint.count(":") < 2:
            # Missing port: append default
            endpoint = endpoint.rstrip("/") + f":{DEFAULT_TCP_PORT}"
        if not endpoint.startswith("tcp://"):
            endpoint = "tcp://" + endpoint
        self._endpoint_private = endpoint
        self.endpoint_edit.setText(endpoint)
        return bind_endpoint_for_public(endpoint)

    def _extract_port(self, endpoint: str) -> Optional[int]:
        if not endpoint:
            return None
        try:
            _, port_str = endpoint.rsplit(":", 1)
            return int(port_str)
        except Exception:
            return None

    def _parse_ssh_target(self, raw: str) -> Tuple[str, int]:
        raw = (raw or "").strip()
        if not raw:
            return "", 22
        if "#" in raw:
            target, port_part = raw.rsplit("#", 1)
            try:
                port = int(port_part)
            except Exception:
                port = 22
            return target or "", port
        return raw, 22

    def _start_reverse_tunnel(self, target: str, ssh_port: int, endpoint: str) -> bool:
        self._stop_tunnel()
        cmd = shutil.which("autossh") or shutil.which("ssh")
        if cmd is None:
            self.status_label.setText("autossh/ssh not found in PATH.")
            return False
        local_port = self._extract_port(endpoint) or DEFAULT_TCP_PORT
        remote_port = local_port
        args = [cmd]
        if cmd.endswith("autossh"):
            args += ["-M", "0"]
        # Use 127.0.0.1 explicitly to avoid IPv6/hostname resolution differences across OSes.
        args += ["-N", "-R", f"{remote_port}:127.0.0.1:{local_port}"]
        if ssh_port != 22:
            args += ["-p", str(ssh_port)]
        args.append(target)
        try:
            self._log_verbose(f"Launching tunnel cmd: {' '.join(args)}")
            self._tunnel_proc = subprocess.Popen(
                args,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                start_new_session=False,
            )
            self.status_label.setText(f"Reverse tunnel via {target} (port {ssh_port})")
            self._log_verbose("Tunnel started.")
            return True
        except Exception as exc:  # noqa: BLE001
            self._tunnel_proc = None
            self.status_label.setText(f"Failed to start tunnel: {exc}")
            self._log_verbose(f"Tunnel start failed: {exc}")
            return False

    def _stop_tunnel(self, *_):
        if self._tunnel_proc is None:
            return
        try:
            self._log_verbose("Stopping tunnel…")
            self._tunnel_proc.terminate()
            self._tunnel_proc.wait(timeout=2)
        except Exception:
            try:
                self._tunnel_proc.kill()
            except Exception:
                pass
        finally:
            self._tunnel_proc = None
            self._log_verbose("Tunnel stopped.")

    def _log_verbose(self, msg: str):
        if self.verbose.isChecked():
            print(f"[napari-stream][receiver] {msg}")

    def _is_running(self) -> bool:
        return self._thread is not None and self._thread.isRunning()

    def _restart_listener(self):
        self._on_stop()
        self._on_start()

    def _on_received(self, arr: np.ndarray, meta: dict):
        name = meta.get("name", "array")
        is_labels = bool(meta.get("is_labels", False))
        self._log_verbose(f"Received data name={name} shape={arr.shape} dtype={arr.dtype} labels={is_labels}")

        # Build kwargs common + per-layer-type
        viewer_kwargs = {}

        # Affine: accept any square >= 2x2 (2x2, 3x3, 4x4, ...)
        if "affine" in meta and not self.ignore_affine.isChecked():
            try:
                A = np.asarray(meta["affine"], dtype=float)
                if A.ndim == 2 and A.shape[0] == A.shape[1] and A.shape[0] >= 2:
                    viewer_kwargs["affine"] = A
            except Exception:
                pass

        # Shared kwargs (supported by both images and labels)
        for key in ("scale", "translate", "opacity", "blending"):
            if key in meta:
                viewer_kwargs[key] = meta[key]

        if is_labels:
            # Labels-specific: do NOT pass image-only args like colormap/contrast_limits/rgb
            layer = self.viewer.add_labels(arr, name=name, **viewer_kwargs)
        else:
            # Image-specific kwargs
            for key in ("colormap", "contrast_limits", "rgb"):
                if key in meta:
                    viewer_kwargs[key] = meta[key]
            layer = self.viewer.add_image(arr, name=name, **viewer_kwargs)

            # Optional autocontrast for grayscale images without provided limits
            if self.autocontrast.isChecked() and "contrast_limits" not in meta and not meta.get("rgb", False):
                try:
                    lo, hi = np.percentile(arr[~np.isnan(arr)], (1, 99))
                    layer.contrast_limits = (float(lo), float(hi))
                except Exception:
                    pass

    def _on_error(self, msg: str):
        self.status_label.setText("Error — see console")
        print("[napari-ipc-bridge] Listener error:\n" + msg)
        traceback.print_exc()


def receiver_widget(viewer=None) -> ReceiverWidget:
    """npe2 command entrypoint: returns a QWidget dock widget.

    Works both when launched from VS Code/example and when starting napari
    from the console, even if type injection doesn't occur. If `viewer`
    is not injected, we try to fetch the current active viewer; as a
    last resort we create a new one.
    """
    try:
        if viewer is None:
            import napari
            viewer = napari.current_viewer() or napari.Viewer()
    except Exception:
        # Extremely defensive: create a viewer if current_viewer failed
        import napari
        viewer = napari.Viewer()
    return ReceiverWidget(viewer)



from __future__ import annotations
import argparse
import json
import os
import time
from typing import Optional, Sequence, Any
from collections.abc import Mapping
import numpy as np
import zmq
from ._utils import default_endpoint


class StreamSender:
    """Send data to a napari receiver.

    Accepts:
      - numpy.ndarray
      - torch.Tensor  (detach().cpu().numpy())
      - blosc2.NDArray
      - zarr.Array
      - Python lists/tuples and dicts (recursively searched for arraylikes;
        nested Python lists of numbers are also converted to NumPy)
    """

    def __init__(
        self,
        endpoint: Optional[str] = None,
        high_water_mark: int = 10,
        linger_ms: int = 2000,
        ensure_delivery: bool = True,
        *,
        verbose: bool = False,
    ):
        if endpoint is None:
            endpoint = os.environ.get("NAPARI_STREAM_ENDPOINT") or default_endpoint()
        self.endpoint = endpoint
        self.ensure_delivery = ensure_delivery
        self.verbose = verbose
        self._ctx = zmq.Context.instance()
        self._sock = self._ctx.socket(zmq.PUSH)
        self._sock.setsockopt(zmq.SNDHWM, high_water_mark)
        self._sock.setsockopt(zmq.LINGER, linger_ms)
        # Fail fast if there is no receiver instead of silently queueing.
        self._sock.setsockopt(zmq.IMMEDIATE, 1)
        self._sock.connect(self.endpoint)

    # ------------------------------ public API ------------------------------

    def send(
        self,
        array: Any,
        *,
        name: Optional[str] = None,
        colormap: Optional[str] = None,
        contrast_limits: Optional[Sequence[float]] = None,
        rgb: Optional[bool] = None,
        affine: Optional[np.ndarray] = None,
        scale: Optional[Sequence[float]] = None,
        translate: Optional[Sequence[float]] = None,
        opacity: Optional[float] = None,
        blending: Optional[str] = None,
        is_labels: bool = False,
    ) -> None:
        """Send one or many arrays.

        If `array` is a list/tuple/dict, recursively find every arraylike leaf and
        send each with a path-qualified name (e.g., `name[0][foo][2]`)."""
        base = name or "array"

        # If the input is a structure, traverse & send each leaf.
        if isinstance(array, Mapping) or (isinstance(array, (list, tuple)) and not isinstance(array, np.ndarray)):
            array = self.retrieve_array_like(base, array).items()
            array = dict(array)
            for path, arr_np in array.items():
                arr_np = self._to_numpy(arr_np)
                self._send_numpy(
                    arr_np,
                    name=path,
                    colormap=colormap,
                    contrast_limits=contrast_limits,
                    rgb=rgb,
                    affine=affine,
                    scale=scale,
                    translate=translate,
                    opacity=opacity,
                    blending=blending,
                    is_labels=is_labels,
                )
            return

        # Single object path
        arr_np = self._to_numpy(array)
        self._send_numpy(
            arr_np,
            name=base,
            colormap=colormap,
            contrast_limits=contrast_limits,
            rgb=rgb,
            affine=affine,
            scale=scale,
            translate=translate,
            opacity=opacity,
            blending=blending,
            is_labels=is_labels,
        )

    def close(self):
        try:
            self._sock.close()
        finally:
            self._sock = None

    # ------------------------------ internals ------------------------------

    def _send_numpy(
        self,
        arr: np.ndarray,
        *,
        name: Optional[str],
        colormap: Optional[str],
        contrast_limits: Optional[Sequence[float]],
        rgb: Optional[bool],
        affine: Optional[np.ndarray] = None,
        scale: Optional[Sequence[float]],
        translate: Optional[Sequence[float]],
        opacity: Optional[float],
        blending: Optional[str],
        is_labels: bool,
    ) -> None:
        # Ensure contiguous so memoryview is a single buffer
        if not (arr.flags["C_CONTIGUOUS"] or arr.flags["F_CONTIGUOUS"]):
            arr = np.ascontiguousarray(arr)
        order = "F" if arr.flags["F_CONTIGUOUS"] else "C"

        meta = {
            "shape": list(arr.shape),
            "dtype": str(arr.dtype),
            "order": order,
            "is_labels": bool(is_labels),
        }
        if name is not None:
            meta["name"] = name
        if colormap is not None:
            meta["colormap"] = colormap
        if contrast_limits is not None:
            meta["contrast_limits"] = list(map(float, contrast_limits))
        if rgb is not None:
            meta["rgb"] = bool(rgb)
        if affine is not None:
            A = np.asarray(affine, dtype=float).tolist()  # (N x N) arbitrary square
            meta["affine"] = A
        if scale is not None:
            meta["scale"] = list(map(float, scale))
        if translate is not None:
            meta["translate"] = list(map(float, translate))
        if opacity is not None:
            meta["opacity"] = float(opacity)
        if blending is not None:
            meta["blending"] = str(blending)

        header = json.dumps(meta).encode("utf-8")
        buf = memoryview(arr)  # zero-copy
        tracker = None
        last_err = None
        if self.verbose:
            print(f"[napari-stream][sender] Trying to send to {self.endpoint}")
        max_attempts = 20
        for attempt in range(1, max_attempts + 1):
            try:
                tracker = self._sock.send_multipart(
                    [header, buf],
                    flags=zmq.NOBLOCK,
                    copy=False,
                    track=self.ensure_delivery,
                )
                last_err = None
                if self.verbose:
                    print(f"[napari-stream][sender] Sending to {self.endpoint}")
                break
            except zmq.Again as e:
                last_err = e
                if self.verbose:
                    print(f"[napari-stream][sender] Receiver not ready (attempt {attempt}/{max_attempts}); retrying…")
                # Allow time for sockets/tunnels to finish connecting
                time.sleep(0.1)
        if last_err is not None:
            raise RuntimeError(f"No receiver available at endpoint {self.endpoint}") from last_err
        if self.ensure_delivery and tracker is not None:
            tracker.wait()
        if self.verbose:
            print(f"[napari-stream][sender] Send complete to {self.endpoint}")

    # ---- conversion helpers ----

    def _to_numpy(self, x: Any) -> np.ndarray:
        """Convert a supported object to np.ndarray without importing optional
        deps unless present. Supports numpy, torch (incl. torchvision.tv_tensors),
        blosc2 NDArray, zarr Array, and numeric Python sequences.
        """
        # Already NumPy
        if isinstance(x, np.ndarray):
            return x

        # --- PyTorch / torchvision.tv_tensors ---
        try:
            import torch  # type: ignore

            if isinstance(x, torch.Tensor):
                return x.detach().cpu().numpy()

            # torchvision.tv_tensors (Image, Mask, Video, BoundingBoxes, etc.)
            mod = getattr(type(x), "__module__", "")
            if mod.startswith("torchvision.tv_tensors"):
                try:
                    t = torch.as_tensor(x)
                except Exception:
                    t = getattr(x, "data", None)
                    if not isinstance(t, torch.Tensor):
                        raise
                return t.detach().cpu().numpy()
        except Exception:
            pass  # torch/torchvision not installed or not applicable

        # --- blosc2.NDArray ---
        try:
            import blosc2  # type: ignore
            if hasattr(blosc2, "NDArray") and isinstance(x, blosc2.NDArray):  # type: ignore[attr-defined]
                return np.asarray(x[:])  # materialize
            to_numpy = getattr(x, "to_numpy", None)
            if callable(to_numpy):
                try:
                    return np.asarray(to_numpy())
                except Exception:
                    pass
        except Exception:
            pass  # blosc2 not installed or not applicable

        # --- zarr arrays (v2/v3) ---
        try:
            import zarr  # type: ignore
            if isinstance(x, getattr(zarr, "Array", ())):
                return np.asarray(x[...])
            core = getattr(zarr, "core", None)
            if core is not None and isinstance(x, getattr(core, "Array", ())):
                return np.asarray(x[...])
        except Exception:
            pass  # zarr not installed or not applicable

        # --- Generic NumPy protocol objects ---
        if hasattr(x, "__array__") or (hasattr(x, "shape") and hasattr(x, "dtype")):
            arr = np.asarray(x)
            if isinstance(arr, np.ndarray) and arr.dtype != object:
                return arr

        # --- Numeric Python sequences (lists/tuples) ---
        if isinstance(x, (list, tuple)):
            try:
                arr = np.asarray(x)
                if isinstance(arr, np.ndarray) and arr.dtype != object:
                    return arr
            except Exception:
                pass

        raise TypeError(f"Unsupported array type: {type(x)!r}")

    def retrieve_array_like(self, name, obj):
        if isinstance(obj, dict):
            new_obj = {}
            for key, value in obj.items():
                result = self.retrieve_array_like(f"{name}[{key}]", value)
                new_obj.update(result)
            return new_obj
        elif isinstance(obj, list):
            new_obj = {}
            for i, value in enumerate(obj):
                result = self.retrieve_array_like(f"{name}[{i}]", value)
                new_obj.update(result)
            return new_obj
        elif self.is_arraylike(obj) and len(obj) >= 1:
            return {name: obj}
        else:
            return {}

    def is_arraylike(self, x):
        """Return True if `x` behaves like a NumPy array (NumPy, Torch, Zarr, Blosc2, etc.)."""
        import numpy as np

        if isinstance(x, (str, bytes, bytearray, dict, set)):
            return False
        if isinstance(x, (np.ndarray, np.generic)):
            return True

        try:
            import torch
            if isinstance(x, torch.Tensor):
                return True
            import torchvision.tv_tensors as tvt
            tv_tensor_types = tuple(
                getattr(tvt, name) for name in dir(tvt)
                if name and name[0].isupper() and hasattr(getattr(tvt, name), "__mro__")
            )
            if isinstance(x, tv_tensor_types):
                return True
        except Exception:
            pass

        try:
            import blosc2
            if hasattr(blosc2, "NDArray") and isinstance(x, blosc2.NDArray):
                return True
        except Exception:
            pass

        try:
            import zarr
            if isinstance(x, zarr.Array):
                return True
        except Exception:
            pass

        if hasattr(x, "__array__") or (hasattr(x, "shape") and hasattr(x, "dtype")):
            return True

        if isinstance(x, (list, tuple)):
            try:
                arr = np.asarray(x)
                return isinstance(arr, np.ndarray) and arr.dtype != object
            except Exception:
                return False

        return False


def send(*args, **kwargs) -> None:
    """Send data to a napari receiver.

    Accepts:
      - numpy.ndarray
      - torch.Tensor  (detach().cpu().numpy())
      - blosc2.NDArray
      - zarr.Array
      - Python lists/tuples and dicts (recursively searched for arraylikes;
        nested Python lists of numbers are also converted to NumPy)
    """
    sender_keys = ("endpoint", "high_water_mark", "linger_ms", "ensure_delivery", "verbose")
    sender_kwargs = {k: kwargs.pop(k) for k in sender_keys if k in kwargs}
    sender = StreamSender(**sender_kwargs)
    sender.send(*args, **kwargs)


def _load_array(path: str, key: Optional[str] = None) -> tuple[np.ndarray, Optional[np.ndarray]]:
    lower = path.lower()
    is_medvol_candidate = lower.endswith((".nii", ".nii.gz", ".nrrd"))
    if is_medvol_candidate:
        from medvol import MedVol  # type: ignore
        mv = MedVol(path)
        return np.asarray(mv.array), np.asarray(mv.affine) if mv.affine is not None else None

    arr = np.load(path)
    if isinstance(arr, np.lib.npyio.NpzFile):
        keys = list(arr.keys())
        if key is None:
            if len(keys) != 1:
                raise ValueError(f"NPZ contains multiple arrays {keys}; specify --key.")
            key = keys[0]
        if key not in keys:
            raise ValueError(f"Key {key!r} not found in NPZ; available: {keys}")
        return np.asarray(arr[key]), None
    return np.asarray(arr), None


def _build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="Send an array to a napari-stream receiver.")
    parser.add_argument("path", help="Path to .npy/.npz or medical volume (.nii/.nii.gz/.nrrd).")
    parser.add_argument("--key", help="Array key inside .npz (if omitted and multiple arrays exist, will error).")
    parser.add_argument("--endpoint", help="Receiver endpoint (tcp://host:port). Falls back to NAPARI_STREAM_ENDPOINT or default.")
    parser.add_argument("--name", help="Layer name.")
    parser.add_argument("--colormap", help="Colormap name.")
    parser.add_argument("--contrast-limits", nargs=2, type=float, metavar=("LOW", "HIGH"), help="Contrast limits.")
    parser.add_argument("--rgb", action="store_true", help="Treat array as RGB.")
    parser.add_argument("--affine", help="Path to .npy/.npz containing affine matrix (overrides MedVol affine).")
    parser.add_argument("--scale", nargs="+", type=float, help="Scale factors.")
    parser.add_argument("--translate", nargs="+", type=float, help="Translation.")
    parser.add_argument("--opacity", type=float, help="Opacity.")
    parser.add_argument("--blending", help="Blending mode.")
    parser.add_argument("--labels", action="store_true", help="Send as labels layer.")
    parser.add_argument("--high-water-mark", type=int, default=10, help="Sender high water mark (default: 10).")
    parser.add_argument("--linger-ms", type=int, default=2000, help="Socket linger in ms (default: 2000).")
    parser.add_argument("--no-ensure-delivery", dest="ensure_delivery", action="store_false", help="Disable delivery tracking.")
    parser.add_argument("--verbose", "-v", action="store_true", help="Verbose logging to stdout.")
    parser.set_defaults(ensure_delivery=True)
    return parser


def main(argv: Optional[Sequence[str]] = None) -> None:
    parser = _build_parser()
    args = parser.parse_args(argv)

    if args.verbose:
        print(f"[napari-stream][sender] Loading data from {args.path}")
    arr, inferred_affine = _load_array(args.path, args.key)
    if args.verbose:
        print(f"[napari-stream][sender] Loaded data shape={arr.shape} dtype={arr.dtype}")
    affine = None
    if args.affine:
        affine, _ = _load_array(args.affine)
    elif inferred_affine is not None:
        affine = inferred_affine

    sender_kwargs = {
        "endpoint": args.endpoint,
        "high_water_mark": args.high_water_mark,
        "linger_ms": args.linger_ms,
        "ensure_delivery": args.ensure_delivery,
        "verbose": args.verbose,
    }

    send_kwargs = {
        "name": args.name,
        "colormap": args.colormap,
        "contrast_limits": args.contrast_limits,
        "rgb": True if args.rgb else None,
        "affine": affine,
        "scale": args.scale,
        "translate": args.translate,
        "opacity": args.opacity,
        "blending": args.blending,
        "is_labels": args.labels,
    }
    send_kwargs = {k: v for k, v in send_kwargs.items() if v is not None}
    if args.verbose:
        print(f"[napari-stream][sender] Attempting to send to {sender_kwargs['endpoint'] or 'default endpoint'}")
    send(arr, **send_kwargs, **sender_kwargs)
    if args.verbose:
        print("[napari-stream][sender] Finished sending.")


if __name__ == "__main__":
    main()
